# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import sys
import typing as ta

from ..lite.timeouts import TimeoutLike
from .base import BaseSubprocesses
from .run import SubprocessRun
from .run import SubprocessRunOutput


##


class AbstractAsyncSubprocesses(BaseSubprocesses):
    @abc.abstractmethod
    async def run_(self, run: SubprocessRun) -> SubprocessRunOutput:
        raise NotImplementedError

    def run(
            self,
            *cmd: str,
            input: ta.Any = None,  # noqa
            timeout: TimeoutLike = None,
            check: bool = False,
            capture_output: ta.Optional[bool] = None,
            **kwargs: ta.Any,
    ) -> ta.Awaitable[SubprocessRunOutput]:
        return self.run_(SubprocessRun(
            cmd=cmd,
            input=input,
            timeout=timeout,
            check=check,
            capture_output=capture_output,
            kwargs=kwargs,
        ))

    #

    @abc.abstractmethod
    async def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        raise NotImplementedError

    #

    async def check_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> str:
        return (await self.check_output(*cmd, **kwargs)).decode().strip()

    #

    async def try_call(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bool:
        if isinstance(await self.async_try_fn(self.check_call, *cmd, **kwargs), Exception):
            return False
        else:
            return True

    async def try_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[bytes]:
        if isinstance(ret := await self.async_try_fn(self.check_output, *cmd, **kwargs), Exception):
            return None
        else:
            return ret

    async def try_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[str]:
        if (ret := await self.try_output(*cmd, **kwargs)) is None:
            return None
        else:
            return ret.decode().strip()



import abc
import contextlib
import typing as ta

from ... import lang
from .base import Kv
from .wrappers import underlying_of


KvT = ta.TypeVar('KvT', bound=Kv)


##


class Closeable(lang.Abstract):
    @abc.abstractmethod
    def close(self) -> None:
        pass


def close(root: Kv) -> None:
    for c in underlying_of(root, Closeable):  # type: ignore[type-abstract]
        c.close()


@contextlib.contextmanager
def closing(kv: KvT) -> ta.Iterator[KvT]:
    try:
        yield kv
    finally:
        close(kv)


##


class Flushable(lang.Abstract):
    @abc.abstractmethod
    def flush(self) -> None:
        pass


def flush(root: Kv) -> None:
    for c in underlying_of(root, Flushable):  # type: ignore[type-abstract]
        c.flush()



import typing as ta

from ... import check
from .base import Kv
from .base import MutableKv
from .wrappers import WrapperKv


K = ta.TypeVar('K')
V = ta.TypeVar('V')

# 'Above' the wrapper
KA = ta.TypeVar('KA')
VA = ta.TypeVar('VA')

# 'Below' the wrapper
KB = ta.TypeVar('KB')
VB = ta.TypeVar('VB')


##


class KeyTransformedKv(WrapperKv[KA, V], ta.Generic[KA, KB, V]):
    def __init__(
            self,
            u: Kv[KB, V],
            *,
            a_to_b: ta.Callable[[KA], KB] | None = None,
            b_to_a: ta.Callable[[KB], KA] | None = None,
    ) -> None:
        super().__init__()

        self._u = u
        self._a_to_b = a_to_b
        self._b_to_a = b_to_a

    def underlying(self) -> ta.Sequence[Kv]:
        return [self._u]

    def __getitem__(self, k: KA, /) -> V:
        fn = check.not_none(self._a_to_b)
        return self._u[fn(k)]

    def __len__(self) -> int:
        return len(self._u)

    def items(self) -> ta.Iterator[tuple[KA, V]]:
        fn = check.not_none(self._b_to_a)
        return ((fn(k), v) for k, v in self._u.items())


class KeyTransformedMutableKey(KeyTransformedKv[KA, KB, V], MutableKv[KA, V], ta.Generic[KA, KB, V]):
    def __init__(
            self,
            u: MutableKv[KB, V],
            *,
            a_to_b: ta.Callable[[KA], KB] | None = None,
            b_to_a: ta.Callable[[KB], KA] | None = None,
    ) -> None:
        super().__init__(
            u,
            a_to_b=a_to_b,
            b_to_a=b_to_a,
        )

    _u: MutableKv[KB, V]

    def __setitem__(self, k: KA, v: V, /) -> None:
        fn = check.not_none(self._a_to_b)
        self._u[fn(k)] = v

    def __delitem__(self, k: KA, /) -> None:
        fn = check.not_none(self._a_to_b)
        del self._u[fn(k)]


##


class ValueTransformedKv(WrapperKv[K, VA], ta.Generic[K, VA, VB]):
    def __init__(
            self,
            u: Kv[K, VB],
            b_to_a: ta.Callable[[VB], VA] | None = None,
    ) -> None:
        super().__init__()

        self._u = u
        self._b_to_a = b_to_a

    def underlying(self) -> ta.Sequence[Kv]:
        return [self._u]

    def __getitem__(self, k: K, /) -> VA:
        fn = check.not_none(self._b_to_a)
        return fn(self._u[k])

    def __len__(self) -> int:
        return len(self._u)

    def items(self) -> ta.Iterator[tuple[K, VA]]:
        fn = check.not_none(self._b_to_a)
        return ((k, fn(v)) for k, v in self._u.items())


class ValueTransformedMutableKv(ValueTransformedKv[K, VA, VB], MutableKv[K, VA], ta.Generic[K, VA, VB]):
    def __init__(
            self,
            u: MutableKv[K, VB],
            *,
            b_to_a: ta.Callable[[VB], VA] | None = None,
            a_to_b: ta.Callable[[VA], VB] | None = None,
    ) -> None:
        super().__init__(u, b_to_a)

        self._a_to_b = a_to_b

    _u: MutableKv[K, VB]

    def __setitem__(self, k: K, v: VA, /) -> None:
        fn = check.not_none(self._a_to_b)
        self._u[k] = fn(v)

    def __delitem__(self, k: K, /) -> None:
        del self._u[k]



"""
TODO:
 - inheritance
 - dynamic registration
 - dynamic switching (skip docker if not running, skip online if not online, ...)
"""
import dataclasses as dc
import enum
import typing as ta

import pytest

from .... import check
from .... import collections as col
from ....docker import all as docker
from ._registry import register


Configable: ta.TypeAlias = pytest.FixtureRequest | pytest.Config


##


class SwitchState(enum.Enum):
    ENABLED = enum.auto()
    DISABLED = enum.auto()
    ONLY = enum.auto()
    IF_SINGLE = enum.auto()


@dc.dataclass(frozen=True, eq=False)
class Switch:
    name: str
    _default_state: SwitchState | ta.Callable[[pytest.Session], SwitchState]

    _: dc.KW_ONLY

    add_marks: ta.Sequence[ta.Any] | None = None

    def default_state(self, session: pytest.Session) -> SwitchState:
        if isinstance(e := self._default_state, SwitchState):
            return e
        elif callable(e):
            return check.isinstance(e(session), SwitchState)
        else:
            raise TypeError(e)

    @property
    def attr(self) -> str:
        return self.name.replace('-', '_')


SWITCHES: ta.Sequence[Switch] = [
    Switch(
        'name',
        lambda _: SwitchState.ENABLED if docker.has_cli() else SwitchState.DISABLED,
    ),

    Switch(
        'docker-guest',
        lambda _: SwitchState.ENABLED if docker.is_likely_in_docker() else SwitchState.DISABLED,
    ),

    Switch(
        'online',
        SwitchState.ENABLED,
    ),

    Switch(
        'integration',
        SwitchState.ENABLED,
    ),

    Switch(
        'high-mem',
        SwitchState.ENABLED,
        add_marks=[
            # https://pytest-xdist.readthedocs.io/en/latest/distribution.html
            pytest.mark.xdist_group('high-mem'),
            pytest.mark.gc_collect_after(),
        ],
    ),

    Switch(
        'slow',
        SwitchState.IF_SINGLE,
    ),
]


SWITCHES_BY_NAME: ta.Mapping[str, Switch] = col.make_map_by(lambda sw: sw.name, SWITCHES, strict=True)
SWITCHES_BY_ATTR: ta.Mapping[str, Switch] = col.make_map_by(lambda sw: sw.attr, SWITCHES, strict=True)


##


SWITCH_STATE_OPT_PREFIXES: ta.Mapping[SwitchState, str] = {
    SwitchState.ENABLED: '--',
    SwitchState.DISABLED: '--no-',
    SwitchState.ONLY: '--only-',
}


def _get_obj_config(obj: Configable) -> pytest.Config:
    if isinstance(obj, pytest.Config):
        return obj
    elif isinstance(obj, pytest.FixtureRequest):
        return obj.config
    else:
        raise TypeError(obj)


def is_disabled(obj: Configable | None, name: str) -> bool:
    check.isinstance(name, str)
    check.in_(name, SWITCHES_BY_NAME)
    return obj is not None and _get_obj_config(obj).getoption(f'--no-{name}')


def skip_if_disabled(obj: Configable | None, name: str) -> None:
    if is_disabled(obj, name):
        pytest.skip(f'{name} disabled')


def get_specified_switches(obj: Configable) -> dict[Switch, SwitchState]:
    ret: dict[Switch, SwitchState] = {}
    for sw in SWITCHES:
        sts = {
            st
            for st, pfx in SWITCH_STATE_OPT_PREFIXES.items()
            if _get_obj_config(obj).getoption(pfx + sw.name)
        }
        if sts:
            if len(sts) > 1:
                raise Exception(f'Multiple switches specified for {sw.name}')
            ret[sw] = check.single(sts)
    return ret


@dc.dataclass(frozen=True)
class ItemSwitches:
    switches: frozenset[Switch]
    not_switches: frozenset[Switch]


def get_item_switches(item: pytest.Item) -> ItemSwitches:
    return ItemSwitches(
        frozenset(sw for sw in SWITCHES if sw.attr in item.keywords),
        frozenset(sw for sw in SWITCHES if ('not_' + sw.attr) in item.keywords),
    )


@register
class SwitchesPlugin:
    def pytest_addoption(self, parser):
        for sw in SWITCHES:
            parser.addoption(f'--no-{sw.name}', action='store_true', default=False, help=f'disable {sw.name} tests')
            parser.addoption(f'--{sw.name}', action='store_true', default=False, help=f'enables {sw.name} tests')
            parser.addoption(f'--only-{sw.name}', action='store_true', default=False, help=f'enables only {sw.name} tests')  # noqa

    def pytest_configure(self, config):
        for sw in SWITCHES:
            config.addinivalue_line('markers', f'{sw.attr}: mark test as {sw.attr}')
            config.addinivalue_line('markers', f'not_{sw.attr}: mark test as not {sw.attr}')

    class _States:
        def __init__(self, session: pytest.Session) -> None:
            super().__init__()

            self.session = session

            self.default_states_by_switch: dict[Switch, SwitchState] = {
                sw: sw.default_state(session)
                for sw in SWITCHES
            }

            self.states_by_switch: dict[Switch, SwitchState] = {
                **self.default_states_by_switch,
                **get_specified_switches(session.config),
            }

            self.switch_sets_by_state: dict[SwitchState, frozenset[Switch]] = {
                st: frozenset(sws)
                for st, sws in col.multi_map(
                    (st, sw)
                    for sw, st in self.states_by_switch.items()
                ).items()
            }

    _states_key: ta.ClassVar[pytest.StashKey[_States]] = pytest.StashKey[_States]()

    def pytest_sessionstart(self, session):
        session.stash[self._states_key] = self._States(session)

    def pytest_collection_modifyitems(self, config, items):
        def process_item(item):
            state: SwitchesPlugin._States = item.session.stash[self._states_key]

            item_switches = get_item_switches(item)

            for sw in item_switches.switches:
                for mk in sw.add_marks or []:
                    item.add_marker(mk)

            if (only_switches := state.switch_sets_by_state.get(SwitchState.ONLY)) is not None:
                if not any(sw in only_switches for sw in item_switches.switches):
                    item.add_marker(pytest.mark.skip(
                        reason=f'skipping switches ({tuple(sw.name for sw in item_switches.switches)!r} (only)',
                    ))
                    return

            disabled_switches = state.switch_sets_by_state.get(SwitchState.DISABLED, frozenset())
            for sw in item_switches.switches:
                if sw in disabled_switches:
                    item.add_marker(pytest.mark.skip(reason=f'skipping switch {sw.name!r} (disabled)'))

            enabled_switches = state.switch_sets_by_state.get(SwitchState.ENABLED, frozenset())
            for sw in item_switches.not_switches:
                if sw in enabled_switches:
                    item.add_marker(pytest.mark.skip(reason=f'skipping switch {sw.name!r} (not-enabled)'))

        for item in items:
            process_item(item)

    def pytest_collection_finish(self, session):
        state: SwitchesPlugin._States = session.stash[self._states_key]

        if_single_switches = state.switch_sets_by_state.get(SwitchState.IF_SINGLE, frozenset())

        if len(session.items) > 1:
            for item in session.items:
                item_switches = get_item_switches(item)
                for sw in item_switches.switches:
                    if sw in if_single_switches:
                        item.add_marker(pytest.mark.skip(reason=f'skipping switch {sw.name!r} (not-single)'))



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import functools
import typing as ta

from ...lite.check import check
from .rewriting import ConfigRewriter
from .rewriting import ConfigRewriterPath


T = ta.TypeVar('T')


##


class MatchingConfigRewriter(ConfigRewriter):
    def __init__(
            self,
            fn: ta.Callable[[ta.Any], ta.Any],
            paths: ta.Iterable[ConfigRewriterPath],
            *,
            recurse: bool = False,
    ) -> None:
        super().__init__()

        self._fn = fn
        self._paths = frozenset(check.isinstance(p, tuple) for p in paths)
        self._recurse = recurse

    def match_path(self, path: ConfigRewriterPath) -> bool:
        for test in self._paths:
            if len(test) != len(path):
                continue
            if all(t is None or p == t for p, t in zip(path, test)):
                return True
        return False

    def rewrite(self, ctx: ConfigRewriter.Context[T]) -> T:
        if self.match_path(ctx.path):
            no = self._fn(ctx.obj)
            if not self._recurse:
                return no
            ctx = dc.replace(ctx, obj=no)

        return super().rewrite(ctx)


def matched_config_rewrite(
        fn: ta.Callable[[ta.Any], ta.Any],
        obj: T,
        *paths: ConfigRewriterPath,
        recurse: bool = False,
        **kwargs: ta.Any,
) -> T:
    return MatchingConfigRewriter(
        functools.partial(fn, **kwargs),
        paths,
        recurse=recurse,
    )(obj)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
  - git var GIT_EDITOR ?
"""
import dataclasses as dc
import os
import tempfile
import typing as ta

from ..lite.cached import cached_nullary
from ..lite.contextmanagers import ExitStacked
from ..lite.contextmanagers import defer
from .run import SubprocessRun
from .run import SubprocessRunnable
from .run import SubprocessRunOutput


##


DEFAULT_USER_TEXT_EDITOR = 'vi'


def get_user_text_editor(default: ta.Optional[str] = None) -> str:
    if default is None:
        default = DEFAULT_USER_TEXT_EDITOR
    return os.environ.get('EDITOR', default)


##


class EditTextWithUserEditor(SubprocessRunnable, ExitStacked):
    def __init__(self, initial_text: str = '') -> None:
        super().__init__()

        self._initial_text = initial_text

    @dc.dataclass(frozen=True)
    class _TempFile:
        path: str
        before_mtime: float

    @cached_nullary
    def _temp_file(self) -> _TempFile:
        with tempfile.NamedTemporaryFile(
                mode='r+',
                delete=False,
        ) as ntf:
            ntf.write(self._initial_text)
            ntf.flush()
            before_mtime = os.stat(ntf.name).st_mtime

        self._enter_context(defer(os.unlink, ntf.name))  # noqa

        return EditTextWithUserEditor._TempFile(
            ntf.name,
            before_mtime,
        )

    def make_run(self) -> SubprocessRun:
        tf = self._temp_file()
        ed = get_user_text_editor()

        return SubprocessRun.of(
            ed,
            tf.path,
        )

    def handle_run_output(self, output: SubprocessRunOutput) -> ta.Optional[str]:
        tf = self._temp_file()

        if output.returncode != 0:
            return None

        after_mtime = os.stat(tf.path).st_mtime

        if after_mtime == tf.before_mtime:
            # No changes made; assuming cancel.
            return None

        with open(tf.path) as f:
            content = f.read().strip()
            if not content:
                # Empty file; assuming cancel.
                return None

        return content


def edit_text_with_user_editor(
        initial_text: str = '',
        subprocesses: ta.Optional[ta.Any] = None,  # AbstractSubprocesses
) -> ta.Optional[str]:
    with EditTextWithUserEditor(initial_text) as cmd:
        return cmd.run(subprocesses)


##


if __name__ == '__main__':
    def _main() -> None:
        text = edit_text_with_user_editor('# Write your message here.\n')
        if text is not None:
            print('User submitted:\n', text)
        else:
            print('Operation canceled.')

    _main()



# ruff: noqa: UP006 UP007
import io
import typing as ta

from .proxy import AsyncIoProxy
from .proxy import _register_async_io_proxy_cls


SelfT = ta.TypeVar('SelfT')


##


@_register_async_io_proxy_cls
class IOBase_AsyncIoProxy(AsyncIoProxy, proxied_cls=io.IOBase):  # noqa
    # https://github.com/python/cpython/blob/e65a1eb93ae35f9fbab1508606e3fbc89123629f/Lib/_pyio.py#L305

    ##
    # Positioning

    async def seek(self, pos, whence=0):
        raise TypeError

    async def tell(self):
        raise TypeError

    async def truncate(self, pos=None):
        raise TypeError

    ##
    # Flush and close

    async def flush(self):
        raise TypeError

    async def close(self):
        raise TypeError

    ##
    # Inquiries

    def seekable(self):
        raise TypeError

    def readable(self):
        raise TypeError

    def writable(self):
        raise TypeError

    @property
    def closed(self):
        raise TypeError

    ##
    # Context manager

    async def __aenter__(self: SelfT) -> SelfT:
        raise TypeError

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        raise TypeError

    ##
    # Lower-level APIs

    def fileno(self):
        raise TypeError

    def isatty(self):
        raise TypeError

    ##
    # Readline[s] and writelines

    async def readline(self, size=-1):
        raise TypeError

    # def __iter__(self):
    #     raise TypeError

    # def __next__(self):
    #     raise TypeError

    async def readlines(self, hint=None):
        raise TypeError

    async def writelines(self, lines):
        raise TypeError


@_register_async_io_proxy_cls
class RawIOBase_AsyncIoProxy(IOBase_AsyncIoProxy, proxied_cls=io.RawIOBase):  # noqa
    async def read(self, size=-1):
        raise TypeError

    async def readall(self):
        raise TypeError

    async def readinto(self, b):
        raise TypeError

    async def write(self, b):
        raise TypeError


@_register_async_io_proxy_cls
class BufferedIOBase_AsyncIoProxy(IOBase_AsyncIoProxy, proxied_cls=io.BufferedIOBase):  # noqa
    async def read(self, size=-1):
        raise TypeError

    async def read1(self, size=-1):
        raise TypeError

    async def readinto(self, b):
        raise TypeError

    async def readinto1(self, b):
        raise TypeError

    async def write(self, b):
        raise TypeError

    async def detach(self):
        raise TypeError


class BufferedIOMixin_AsyncIoProxy(BufferedIOBase_AsyncIoProxy):  # noqa
    @property
    def raw(self):
        raise TypeError

    @property
    def name(self):
        raise TypeError

    @property
    def mode(self):
        raise TypeError


@_register_async_io_proxy_cls
class BytesIO_AsyncIoProxy(BufferedIOBase_AsyncIoProxy, proxied_cls=io.BytesIO):  # noqa
    async def getvalue(self):
        raise TypeError

    async def getbuffer(self):
        raise TypeError


@_register_async_io_proxy_cls
class BufferedReader_AsyncIoProxy(BufferedIOMixin_AsyncIoProxy, proxied_cls=io.BufferedReader):  # noqa
    async def peek(self, size=0):
        raise TypeError


@_register_async_io_proxy_cls
class BufferedWriter_AsyncIoProxy(BufferedIOMixin_AsyncIoProxy, proxied_cls=io.BufferedWriter):  # noqa
    pass


@_register_async_io_proxy_cls
class BufferedRWPair_AsyncIoProxy(BufferedIOBase_AsyncIoProxy, proxied_cls=io.BufferedRWPair):  # noqa
    async def peek(self, size=0):
        raise TypeError


@_register_async_io_proxy_cls
class BufferedRandom_AsyncIoProxy(BufferedWriter_AsyncIoProxy, BufferedReader_AsyncIoProxy, proxied_cls=io.BufferedRandom):  # noqa
    pass


@_register_async_io_proxy_cls
class FileIO_AsyncIoProxy(RawIOBase_AsyncIoProxy, proxied_cls=io.FileIO):  # noqa
    @property
    def closefd(self):
        raise TypeError

    @property
    def mode(self):
        raise TypeError


@_register_async_io_proxy_cls
class TextIOBase_AsyncIoProxy(IOBase_AsyncIoProxy, proxied_cls=io.TextIOBase):  # noqa
    async def read(self, size=-1):
        raise TypeError

    async def write(self, s):
        raise TypeError

    async def detach(self):
        raise TypeError

    @property
    def encoding(self):
        raise TypeError

    @property
    def newlines(self):
        raise TypeError

    @property
    def errors(self):
        raise TypeError


@_register_async_io_proxy_cls
class TextIOWrapper_AsyncIoProxy(TextIOBase_AsyncIoProxy, proxied_cls=io.TextIOWrapper):  # noqa
    @property
    def line_buffering(self):
        raise TypeError

    @property
    def write_through(self):
        raise TypeError

    @property
    def buffer(self):
        raise TypeError

    async def reconfigure(
            self,
            *,
            encoding=None,
            errors=None,
            newline=Ellipsis,
            line_buffering=None,
            write_through=None,
    ):
        raise TypeError

    @property
    def name(self):
        raise TypeError


@_register_async_io_proxy_cls
class StringIO_AsyncIoProxy(TextIOWrapper_AsyncIoProxy, proxied_cls=io.StringIO):  # noqa
    async def getvalue(self):
        raise TypeError



import typing as ta

from ... import check
from ... import lang


##


@lang.cached_function
def _cyclic_dependency_proxy() -> tuple[type, ta.Callable[[ta.Any, ta.Any], None]]:
    import wrapt  # noqa

    class _CyclicDependencyPlaceholder(lang.Final):

        def __init__(self, cls: ta.Any) -> None:
            super().__init__()
            self.__cls = cls

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.__cls!r})'

    class _CyclicDependencyProxy(wrapt.ObjectProxy, lang.Final):  # noqa

        def __init__(self, cls):
            super().__init__(_CyclicDependencyPlaceholder(cls))
            if isinstance(cls, type):
                self._self__class__ = cls  # noqa

        def __repr__(self) -> str:
            return f'_CyclicDependencyProxy({self.__wrapped__!r})'

        def __getattribute__(self, att):
            if att == '__class__':
                try:
                    return object.__getattribute__(self, '_self__class__')  # noqa
                except AttributeError:
                    pass
            return object.__getattribute__(self, att)  # noqa

    def set_obj(prox, obj):
        check.state(type(prox) is _CyclicDependencyProxy)
        check.not_isinstance(obj, _CyclicDependencyPlaceholder)
        check.isinstance(prox.__wrapped__, _CyclicDependencyPlaceholder)
        if hasattr(prox, '_self__class__'):
            check.issubclass(type(obj), prox._self__class__)  # noqa
        prox.__wrapped__ = obj
        if hasattr(prox, '_self__class__'):
            del prox._self__class__  # noqa

    return (_CyclicDependencyProxy, set_obj)  # noqa



import typing as ta

from ... import dataclasses as dc
from .metadata import FieldMetadata
from .metadata import ObjectMetadata


T = ta.TypeVar('T')


##


def with_field_metadata(**kwargs: ta.Any) -> dc.field_modifier:
    @dc.field_modifier
    def inner(f: dc.Field) -> dc.Field:
        return dc.set_field_metadata(f, {
            FieldMetadata: f.metadata.get(FieldMetadata, FieldMetadata()).update(**kwargs),
        })
    return inner


def update_fields_metadata(
        fields: ta.Iterable[str] | None = None,
        **kwargs: ta.Any,
) -> ta.Callable[[type[T]], type[T]]:
    def inner(a: str, f: dc.Field) -> dc.Field:
        return dc.set_field_metadata(f, {
            FieldMetadata: f.metadata.get(FieldMetadata, FieldMetadata()).update(**kwargs),
        })

    return dc.update_fields(inner, fields)


def update_object_metadata(
        cls: type | None = None,
        **kwargs: ta.Any,
):
    def inner(cls):
        return dc.append_class_metadata(cls, ObjectMetadata(**kwargs))

    if cls is not None:
        inner(cls)
        return cls
    else:
        return inner
